use utf8;
use strict;
use warnings FATAL => 'all';

package Xyzzy::Request;

use Spiffy -Base;

use CGI::Fast;
use CGI::Util;

use Xyzzy::Header;
use Xyzzy::Status;

field 'outheaders';
field 'config';
field 'content';
field 'type';
field 'charset';
field 'cgi';

sub canonicalheader() {
	my $key = $_[0];
	die "Invalid header name '$key'\n"
		unless Xyzzy::Header::istoken($key);
	$key =~ s/([a-z]+)/ucfirst(lc($1))/gei;
	return $key;
}

sub addheader {
	my $key = canonicalheader(shift);

	my $headers = $self->outheaders;
	my $header = $headers->{$key};
	$headers->{$key} = $header = []
		unless defined $header;

	push @$header, shift
		while ref $_[0];

	push @$header, new Xyzzy::Header(@_)
		if @_;
}

sub setheader {
	my $key = canonicalheader(shift);

	if(@_ == 1 && !defined $_[0]) {
		delete $self->outheaders->{$key};
		return;
	}

	my @header;

	push @header, shift
		while ref $_[0];

	push @header, new Xyzzy::Header(@_)
		if @_;

	$self->outheaders->{$key} = \@header;
}

sub secure {
	return lc($self->https) eq 'on';
}

sub setcookie {
	my $name = shift;
	my $value = CGI::Util::escape(shift);
	my %defaults = (
			Path => $self->url(-absolute => 1),
			Version => 1,
		);
	undef $defaults{secure}
		if $self->secure;
	my @args;
	while(@_) {
		my $key = canonicalheader(shift);
		my $val = shift;
		delete $defaults{$key};
		if($key eq 'Secure') {
			next if defined $val && !$val;
			undef $val;
		}
		push @args, $key, $val;
	}
	while(my ($key, $val) = each(%defaults)) {
		push @args, canonicalheader($key), $val;
	}

	$self->addheader('Set-Cookie', $name => $value, @args);
}

sub status {
	return $self->{status} unless @_;
	return $self->{status} = $_[0] if ref $_[0];
	return $self->{status} = new Xyzzy::Status(@_);
}

sub mimetype {
	my $mimetype = lc(shift);
	my $charset;
	my @args = @_;
	while(@args) {
		my $key = shift @args;
		my $val = shift @args;
		$charset = $val
			if defined $key && defined $val && lc($key) eq 'charset';
	}
	$self->type($mimetype);
	$self->charset($charset);
	$self->setheader('Content-Type', $mimetype, @_);
}

sub new {
	my $config = shift;
	my $cgi = new CGI::Fast or return;
	my $headers = {
		'Pragma' => [new Xyzzy::Header('no-cache')],
		'Cache-Control' => [
				new Xyzzy::Header('no-cache'),
				new Xyzzy::Header('must-revalidate')
			],
		'Expires' => [new Xyzzy::Header('Thu, 01 Jan 1970 00:00:00 GMT')],
	};
	my $req = super(
			config => $config,
			status => new Xyzzy::Status(200),
			outheaders => $headers,
			cgi => $cgi
		);
	$req->mimetype('text/html', charset => 'UTF-8');
	return $req;
}

our $AUTOLOAD;
sub AUTOLOAD {
	my $sub = $AUTOLOAD;
	my $off = rindex($sub, '::');
	substr($sub, 0, $off + 2) = ''
		if $off != -1;
	return $self->cgi->$sub(@_);
}
